@@ -5,7 +5,7 @@ import grpc
 
 
 class CancellableStreamThread:
-    def __init__(self, callback, create_stream, attribute=None):
+    def __init__(self, callback, create_stream, logger, attribute=None):
         """
         Feeds data from the gRPC stream into the callback
         until the cancel method is called.
@@ -23,6 +23,10 @@ class CancellableStreamThread:
                 except grpc.RpcError as e:
                     if e.code() == grpc.StatusCode.CANCELLED:
                         return
+                    if logger:
+                        logger.warn("Error creating stream %s" % str(e))
+                    if e.code() == grpc.StatusCode.UNAUTHENTICATED:
+                        return
                     # wait a short duration before re-registering when
                     # agent unavailable or other rpc error
                     time.sleep(0.1)
