@@ -25,6 +25,21 @@ def current_timestamp():
     return int(time.time() * 1000)
 
 
+def validate_type(
+    given,  # type: Any
+    expected_types,  # type: List[type]
+):
+    if type(given) not in expected_types:
+        raise TypeError(
+            "Formant client method input '"
+            + str(given)
+            + "' has type "
+            + str(type(given))
+            + ", but expected one of: "
+            + str(expected_types)
+        )
+
+
 class Client:
     """
     A client for interacting with the Formant Agent. There are methods for:
@@ -81,6 +96,9 @@ class Client:
     ):
         # type: (...) -> None
         """Post a text datapoint to a telemetry stream."""
+
+        validate_type(value, [str])
+
         self.agent_stub.PostData(
             datapoint_pb2.Datapoint(
                 stream=stream,
@@ -100,6 +118,9 @@ class Client:
     ):
         # type: (...) -> None
         """Post a JSON datapoint to a telemetry stream."""
+
+        validate_type(value, [str])
+
         self.agent_stub.PostData(
             datapoint_pb2.Datapoint(
                 stream=stream,
@@ -113,12 +134,15 @@ class Client:
     def post_numeric(
         self,
         stream,  # type: str
-        value,  # type: Union[int, float]
+        value,  # type: Union[float, int]
         tags=None,  # type: Optional[Dict[str, str]]
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
         """Post a numeric datapoint to a telemetry stream."""
+
+        validate_type(value, [float, int])
+
         self.agent_stub.PostData(
             datapoint_pb2.Datapoint(
                 stream=stream,
@@ -132,15 +156,22 @@ class Client:
     def post_numericset(
         self,
         stream,  # type: str
-        numerics_dict,  # type: Dict[str, Tuple[Union[int, float], Optional[str]]]
+        numerics_dict,  # type: Dict[str, Tuple[Union[float, int], Optional[str]]]
         tags=None,  # type: Optional[Dict[str, str]]
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
         """Post a numeric set datapoint to a telemetry stream."""
 
+        validate_type(numerics_dict, [dict])
+
         numeric_set = math_pb2.NumericSet()
         for k, v in numerics_dict.items():
+            validate_type(k, [str])
+            validate_type(v, [tuple])
+            validate_type(v[0], [float, int])
+            if v[1]:
+                validate_type(v[1], [str])
             numeric_set.numerics.extend(
                 [
                     math_pb2.NumericSetEntry(
@@ -162,13 +193,16 @@ class Client:
         self,
         stream,  # type: str
         # raw bytes:
-        value,  # type: str
+        value,  # type: bytes
         content_type="image/jpg",  # type: Literal["image/jpg", "image/png"]
         tags=None,  # type: Optional[Dict[str, str]]
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
         """Post an image datapoint to a telemetry stream."""
+
+        validate_type(value, [bytes])
+
         self.agent_stub.PostData(
             datapoint_pb2.Datapoint(
                 stream=stream,
@@ -189,8 +223,12 @@ class Client:
         # type: (...) -> None
         """Post a bitset datapoint to a telemetry stream."""
 
+        validate_type(bitset_dict, [dict])
+
         bitset = math_pb2.Bitset()
         for k, v in bitset_dict.items():
+            validate_type(k, [str])
+            validate_type(v, [bool])
             bitset.bits.extend([math_pb2.Bit(key=k, value=v)])
 
         self.agent_stub.PostData(
@@ -206,13 +244,17 @@ class Client:
     def post_geolocation(
         self,
         stream,  # type: str
-        latitude,  # type: float
-        longitude,  # type: float
+        latitude,  # type: Union[float, int]
+        longitude,  # type: Union[float, int]
         tags=None,  # type: Optional[Dict[str, str]]
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
         """Post a geolocation datapoint to a telemetry stream."""
+
+        for _ in [latitude, longitude]:
+            validate_type(_, [float, int])
+
         self.agent_stub.PostData(
             datapoint_pb2.Datapoint(
                 stream=stream,
@@ -228,15 +270,20 @@ class Client:
     def post_battery(
         self,
         stream,  # type: str
-        percentage,  # type: float
-        voltage=None,  # type: Optional[float]
-        current=None,  # type: Optional[float]
-        charge=None,  # type: Optional[float]
+        percentage,  # type: Union[int, float]
+        voltage=None,  # type: Optional[Union[int, float]]
+        current=None,  # type: Optional[Union[int, float]]
+        charge=None,  # type: Optional[Union[int, float]]
         tags=None,  # type: Optional[Dict[str, str]]
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
         """Post a battery datapoint to a telemetry stream."""
+
+        for _ in [percentage, voltage, current, charge]:
+            if _:
+                validate_type(_, [float, int])
+
         battery = health_pb2.Battery()
         battery.percentage = percentage
         if voltage is not None:
@@ -262,16 +309,22 @@ class Client:
         self,
         parent_frame,  # type: str
         child_frame,  # type: str
-        tx,  # type: float
-        ty,  # type: float
-        tz,  # type: float
-        rx,  # type: float
-        ry,  # type: float
-        rz,  # type: float
-        rw,  # type: float
+        tx,  # type: Union[int, float]
+        ty,  # type: Union[int, float]
+        tz,  # type: Union[int, float]
+        rx,  # type: Union[int, float]
+        ry,  # type: Union[int, float]
+        rz,  # type: Union[int, float]
+        rw,  # type: Union[int, float]
     ):
         # type: (...) -> None
         """Adds a transform frame, used to position datapoints in 3D space."""
+
+        for _ in [parent_frame, child_frame]:
+            validate_type(_, [str])
+        for _ in [tx, ty, tz, rx, ry, rz, rw]:
+            validate_type(_, [float, int])
+
         frame = math_pb2.TransformFrame()
         frame.parent_frame = parent_frame
         frame.child_frame = child_frame
@@ -293,15 +346,30 @@ class Client:
         notify=False,  # type: bool
         tags=None,  # type: Optional[Dict[str, str]]
         timestamp=None,  # type: Optional[int]
+        end_timestamp=None,  # type: Optional[int]
+        severity="info",  # type: Literal["info", "warning", "critical", "error"]
     ):
         """Creates and ingests an event."""
+
+        validate_type(message, [str])
+
+        severity_pb = event_pb2.INFO
+        if severity == "warning":
+            severity_pb = event_pb2.WARNING
+        elif severity == "critical":
+            severity_pb = event_pb2.CRITICAL
+        elif severity == "error":
+            severity_pb = event_pb2.ERROR
+
         self.agent_stub.CreateEvent(
             agent_pb2.CreateEventRequest(
                 event=event_pb2.Event(
                     message=message,
                     notification_enabled=notify,
                     tags=tags if tags else {},
+                    severity=severity_pb,
                     timestamp=timestamp if timestamp else current_timestamp(),
+                    end_timestamp=end_timestamp,
                 )
             )
         )
@@ -362,7 +430,7 @@ class Client:
             return self._get_command_request_stream(command_filter)
 
         self._command_request_callback_streams[f] = CancellableStreamThread(
-            f, create_stream, attribute="request"
+            f, create_stream, self.logger, attribute="request"
         )
 
     def unregister_command_request_callback(
@@ -394,7 +462,7 @@ class Client:
             return self._get_teleop_stream(stream_filter)
 
         self._teleop_callback_streams[f] = CancellableStreamThread(
-            f, create_stream, attribute="control_datapoint"
+            f, create_stream, self.logger, attribute="control_datapoint"
         )
 
     def unregister_teleop_callback(
@@ -425,6 +493,7 @@ class Client:
 
         return self._app_config.get(key, default)
 
+    @handle_agent_exceptions
     def get_config_blob_data(self):
         # type: (...) -> str
         """
@@ -449,6 +518,15 @@ class Client:
         """
         self._config_update_callbacks.remove(f)
 
+    @handle_agent_exceptions
+    def get_teleop_info(self):
+        # type: (...) -> agent_pb2.GetTeleopInfoResponse
+        """
+        Returns current information about teleop connection count.
+        """
+        request = agent_pb2.GetTeleopInfoRequest()
+        return self.agent_stub.GetTeleopInfo(request)
+
     # Internal methods
 
     def _setup_agent_communication(self):
